/* Python RxRPC transport container object
 *
 * Copyright (C) 2015 Red Hat, Inc. All Rights Reserved.
 * Written by David Howells (dhowells@redhat.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public Licence
 * as published by the Free Software Foundation; either version
 * 2 of the Licence, or (at your option) any later version.
 */

#include <Python.h>
#include "structmember.h"
#include <arpa/inet.h>
#include "py_rxgen.h"
#include "rxgen.h"

/*
 * RxRPC transport container.
 */
static int
py_rx_transport_init(PyObject *_self, PyObject *args, PyObject *kwds)
{
	struct py_rx_transport *self = (struct py_rx_transport *)_self;
	self->x = NULL;
	return 0;
}

static void
py_rx_transport_dealloc(struct py_rx_transport *self)
{
	if (self->x) {
		rx_free_transport(self->x);
		self->x = NULL;
	}
	Py_TYPE(self)->tp_free((PyObject *)self);
}

PyTypeObject py_rx_transportType = {
	PyVarObject_HEAD_INIT(NULL, 0)
	"kafs.rx_transport",		/*tp_name*/
	sizeof(struct py_rx_transport),	/*tp_basicsize*/
	0,				/*tp_itemsize*/
	(destructor)py_rx_transport_dealloc, /*tp_dealloc*/
	0,				/*tp_print*/
	0,				/*tp_getattr*/
	0,				/*tp_setattr*/
	0,				/*tp_compare*/
	0,				/*tp_repr*/
	0,				/*tp_as_number*/
	0,				/*tp_as_sequence*/
	0,				/*tp_as_mapping*/
	0,				/*tp_hash */
	0,				/*tp_call*/
	0,				/*tp_str*/
	0,				/*tp_getattro*/
	0,				/*tp_setattro*/
	0,				/*tp_as_buffer*/
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
	"RxRPC transport container",	/* tp_doc */
	0,				/* tp_traverse */
	0,				/* tp_clear */
	0,				/* tp_richcompare */
	0,				/* tp_weaklistoffset */
	0,				/* tp_iter */
	0,				/* tp_iternext */
	0,				/* tp_methods */
	0,				/* tp_members */
	0,				/* tp_getset */
	0,				/* tp_base */
	0,				/* tp_dict */
	0,				/* tp_descr_get */
	0,				/* tp_descr_set */
	0,				/* tp_dictoffset */
	py_rx_transport_init,		/* tp_init */
	0,				/* tp_alloc */
	0,				/* tp_new */
};

/*
 * Set up an RxRPC transport.
 */
PyObject *
kafs_py_rx_new_transport(PyObject *_self, PyObject *args)
{
	struct py_rx_transport *obj;
	struct kafs_transport_handle *trans;
	union {
		struct sockaddr sa;
		struct sockaddr_in sin;
		struct sockaddr_in6 sin6;
	} sa;
	const char *local_address = NULL;
	socklen_t salen;
	uint16_t local_port = 0;

	if (!PyArg_ParseTuple(args, "|Hs", &local_port, &local_address))
		return NULL;

	memset(&sa, 0, sizeof(sa));
	if (!local_address && !local_port) {
		sa.sin.sin_family = AF_INET;
		salen = sizeof(sa.sin);
	} else if (!local_address) {
		sa.sin.sin_family = AF_INET;
		sa.sin.sin_port = htons(local_port);
		salen = sizeof(sa.sin);
	} else if (inet_pton(AF_INET, local_address, &sa.sin.sin_addr)) {
		sa.sin.sin_family = AF_INET;
		sa.sin.sin_port = htons(local_port);
		salen = sizeof(sa.sin);
	} else if (inet_pton(AF_INET6, local_address, &sa.sin.sin_addr)) {
		sa.sin6.sin6_family = AF_INET6;
		sa.sin6.sin6_port = htons(local_port);
		salen = sizeof(sa.sin6);
	} else {
		return PyErr_Format(PyExc_RuntimeError,
				    "Unsupported network address '%s'", local_address);
	}

	obj = (struct py_rx_transport *)_PyObject_New(&py_rx_transportType);
	if (!obj)
		return PyExc_MemoryError;
	py_rx_transport_init((PyObject *)obj, NULL, NULL);
	assert(obj->x == NULL);

	trans = rx_new_transport(&sa.sa, salen);
	if (!trans) {
		Py_DECREF(obj);
		return errno == ENOMEM ? PyExc_MemoryError :
			PyErr_SetFromErrno(PyExc_IOError);
	}
	obj->x = trans;
	return (PyObject *)obj;
}

/*
 * Load a kAFS transport module.
 */
PyObject *
kafs_py_load_kafs_transport(PyObject *_self, PyObject *args)
{
	const char *name = NULL;
	int ret;

	if (!PyArg_ParseTuple(args, "s", &name))
		return NULL;

	ret = load_kafs_transport_module(name);
	if (ret < 0)
		return errno == ENOMEM ? PyExc_MemoryError :
			PyErr_SetFromErrno(PyExc_IOError);
	Py_RETURN_TRUE;
}
