/* Dynamically load the Rx transport module
 *
 * Copyright (C) 2015 Red Hat, Inc. All Rights Reserved.
 * Written by David Howells (dhowells@redhat.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public Licence
 * as published by the Free Software Foundation; either version
 * 2 of the Licence, or (at your option) any later version.
 */

#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <netinet/in.h>
#include "transport.h"
#include "rxgen.h"

#define RXGEN_CALL_MAGIC	0x52584745U
#define RXGEN_BUF_MAGIC		(0x52420000U | __LINE__)
#define RXGEN_BUF_DEAD		(0x6b6bU | __LINE__)

#define debug(fmt, ...) do { if (0) printf("TRAN: "fmt, ## __VA_ARGS__); } while (0)

uint32_t rxgen_dec_padding_sink;

static struct kafs_transport *transport;
static char transport_id[] = kafs_transport_id;

static void rxrpc_data_ready(struct kafs_call_handle *call_handle, void *caller_data);

/*
 * Load a transport module and extract the entry points
 */
int load_kafs_transport_module(const char *name)
{
	void *trans;
	char *err;

	trans = dlopen(name, RTLD_LAZY);
	if (!trans) {
		err = dlerror();
		goto error;
	}

	dlerror(); /* clear error */
	transport = dlsym(trans, "kafs_transport");
	err = dlerror();
	if (err)
		goto error_2;

	if (memcmp(transport->id, transport_id, sizeof(transport_id)) != 0) {
		fprintf(stderr, "Transport '%s' '%s' %zu\n",
			transport->id, transport_id, sizeof(transport_id));
		err = "Transport has incorrect ID string\n";
		goto error_2;
	}
	return 0;

error_2:
	dlclose(trans);
error:
	fprintf(stderr, "%s: %s\n", name, err);
	errno = -ELIBBAD;
	return -1;
}

/*
 * Set up a new transport
 */
struct kafs_transport_handle *rx_new_transport(const struct sockaddr *local_sa,
					       socklen_t local_salen)
{

	if (!transport) {
		fprintf(stderr, "No Rx transport loaded\n");
		errno = EINVAL;
		return NULL;
	}

	return transport->init_transport(local_sa, local_salen);
}

/*
 * Destroy a transport
 */
void rx_free_transport(struct kafs_transport_handle *trans)
{
	transport->free_transport(trans);
}

/*
 * Set up a new connection
 */
struct rx_connection *rx_new_connection(struct kafs_transport_handle *trans_handle,
					const struct sockaddr *sa,
					socklen_t salen,
					uint16_t service,
					int exclusive,
					const char *cellname,
					enum kafs_connection_auth_level level)
{
	struct rx_connection *conn;
	const struct sockaddr_in *sin;
	const struct sockaddr_in6 *sin6;
	struct kafs_security_handle *security;
	
	/* Do some basic parameter checking */
	switch (sa->sa_family) {
	case 0:
		errno = EDESTADDRREQ;
		return NULL;
	case AF_INET:
		if (salen != sizeof(struct sockaddr_in)) {
			errno = EINVAL;
			return NULL;
		}
		sin = (const struct sockaddr_in *)sa;
		if (!sin->sin_port) {
			errno = EDESTADDRREQ;
			return NULL;
		}
		break;
	case AF_INET6:
		if (salen != sizeof(struct sockaddr_in6)) {
			errno = EINVAL;
			return NULL;
		}
		sin6 = (const struct sockaddr_in6 *)sa;
		if (!sin6->sin6_port) {
			errno = EDESTADDRREQ;
			return NULL;
		}
		break;
	default:
		errno = EPROTOTYPE;
		return NULL;
	}

	if (level < kafs_connection_no_auth ||
	    level > kafs_connection_encrypt) {
		errno = EINVAL;
		return NULL;
	}

	/* Set up our connection tracking */
	conn = calloc(1, sizeof(*conn));
	if (!conn)
		return NULL;

	/* Create a security handle */
	security = transport->new_security(trans_handle, cellname, level);
	if (!security)
		goto error_conn;

	/* Open the transport connection */
	conn->handle = transport->open_connection(trans_handle,
						  security,
						  sa, salen,
						  service,
						  exclusive);
	transport->free_security(security);
	if (!conn->handle)
		goto error_conn;

	return conn;

error_conn:
	free(conn);
	return NULL;
}

/*
 * Close an RxRPC client connection.  This will cause all outstanding
 * operations to be aborted by the kernel..
 */
void rx_close_connection(struct rx_connection *conn)
{
	transport->close_connection(conn->handle);
	free(conn);
}

/*
 * Create a call structure with which to make a call.  It is given one buffer
 * ready to go.
 */
struct rx_call *rxrpc_make_call(struct rx_connection *conn)
{
	struct rx_call *call;
	struct rx_buf *buf;
	uint8_t *data;

	call = calloc(1, sizeof(struct rx_call));
	if (!call)
		goto error_0;

	buf = calloc(1, sizeof(struct rx_buf));
	if (!buf)
		goto error_1;

	data = malloc(RXGEN_BUFFER_SIZE);
	if (!data)
		goto error_2;

	buf->magic = RXGEN_BUF_MAGIC;
	buf->buf = data;

	call->magic = RXGEN_CALL_MAGIC;
	call->conn = conn;
	call->state = rx_call_cl_not_started;
	call->more_send = true;
	call->more_recv = true;
	call->buffer_head = buf;
	call->buffer_tail = buf;
	call->data_start = data;
	call->data_cursor = data;
	call->data_stop = data + RXGEN_BUFFER_SIZE;
	call->buffer_space = RXGEN_BUFFER_SIZE;

	call->handle = transport->make_call(conn->handle, rxrpc_data_ready, call);
	if (!call->handle)
		goto error_3;

	debug("Alloc: buf=%p data=%p\n", buf, data);
	return call;

error_3:
	free(data);
error_2:
	free(buf);
error_1:
	free(call);
error_0:
	return NULL;
}

/*
 * Create a call structure with which to accept a call.  It is given one buffer
 * ready to go.
 */
#if 0
struct rx_call *rxrpc_accept_call(struct kafs_service_handle *service_handle,
				  void *caller_data,
				  const struct sockaddr *peer,
				  socklen_t peerlen)
{
	struct rx_service *service = caller_data;
	struct rx_call *call;
	struct rx_buf *buf;
	uint8_t *data;

	call = calloc(1, sizeof(struct rx_call));
	if (!call)
		goto error_0;

	buf = calloc(1, sizeof(struct rx_buf));
	if (!buf)
		goto error_1;

	data = malloc(RXGEN_BUFFER_SIZE);
	if (!data)
		goto error_2;

	buf->magic = RXGEN_BUF_MAGIC;
	buf->buf = data;

	call->magic = RXGEN_CALL_MAGIC;
	call->conn = conn;
	call->state = rx_call_sv_not_started;
	call->more_send = true;
	call->more_recv = true;
	call->buffer_head = buf;
	call->buffer_tail = buf;
	call->data_start = data;
	call->data_cursor = data;
	call->data_stop = data;
	call->buffer_space = RXGEN_BUFFER_SIZE;

	call->handle = transport->accept_call(service_handle, peer, peerlen);
	if (!call->handle)
		goto error_3;

	debug("Alloc: buf=%p data=%p\n", buf, data);
	return call;

error_3:
	free(data);
error_2:
	free(buf);
error_1:
	free(call);
error_0:
	return NULL;
}
#endif

static void rxrpc_check_buf(const struct rx_buf *buf)
{
	if (0) {
		if ((buf->magic ^ RXGEN_BUF_MAGIC) & 0xffff0000U)
			abort();
		if (buf->io_cursor > RXGEN_BUFFER_SIZE)
			abort();
	}
}

static void rxrpc_check_call(const struct rx_call *call)
{
	if (0) {
		const struct rx_buf *cursor;

		if (call->magic != RXGEN_CALL_MAGIC)
			abort();

		for (cursor = call->buffer_head; cursor != call->buffer_tail; cursor = cursor->next)
			rxrpc_check_buf(cursor);
		rxrpc_check_buf(cursor);
		if (cursor->next)
			abort();
	}
}

/*
 * Send buffered data.
 */
int rxrpc_send_data(struct rx_call *call)
{
	struct rx_buf *cursor;
	struct iovec iov[16];
	unsigned more;
	int ioc, ret;

	debug("-->rxrpc_send_data(%u,0x%x)\n", call->state, call->data_count);

	rxrpc_check_call(call);

	/* Switch into encode state */
	switch (call->state) {
	case rx_call_cl_not_started:
	case rx_call_sv_processing:
		call->state++;
	case rx_call_cl_encoding_params:
	case rx_call_sv_encoding_response:
		break;
	default:
		fprintf(stderr, "RxRPC: Send in bad call state (%d)\n", call->state);
		abort();
	}

more_to_send:
	/* We may have a completely sent buffer on the front of the queue if
	 * this was the last buffer on the last send.  The buffer queue isn't
	 * allowed to be empty at any point.
	 */
	cursor = call->buffer_head;
	if (cursor->io_cursor == RXGEN_BUFFER_SIZE) {
		struct rx_buf *sent = cursor;
		if (sent == call->buffer_tail)
			abort();
		call->buffer_head = cursor = sent->next;
		sent->magic = RXGEN_BUF_DEAD;
		free(sent->buf);
		free(sent);
	}

	more = 1;
	for (ioc = 0; ioc < 16; ioc++) {
		rxrpc_check_buf(cursor);

		unsigned io_cursor = cursor->io_cursor;
		unsigned end = RXGEN_BUFFER_SIZE;

		if (io_cursor == RXGEN_BUFFER_SIZE)
			abort();
		if (cursor == call->buffer_tail)
			end = call->data_cursor - cursor->buf;

		debug("BUF[%02u] %04x %04x\n", ioc, io_cursor, end);

		iov[ioc].iov_base = cursor->buf + io_cursor;
		iov[ioc].iov_len = end - io_cursor;
		if (cursor == call->buffer_tail) {
			if (!call->more_send)
				more = 0;
			ioc++;
			break;
		}
		cursor = cursor->next;
	}

	ret = transport->send_data(call->handle, iov, ioc, more);
	if (ret == -1)
		return -1;

	call->bytes_sent += ret;
	call->data_count -= ret;
	call->known_to_kernel = 1;

	/* Free up any completely sent buffers, without completely emptying the
	 * queue.
	 */
	cursor = call->buffer_head;
	do {
		struct rx_buf *sent = cursor;
		unsigned int count = RXGEN_BUFFER_SIZE - cursor->io_cursor;

		if (count > (unsigned int)ret)
			count = (unsigned int)ret;
		cursor->io_cursor += count;
		ret -= count;
		if (cursor == call->buffer_tail)
			break;
		sent = cursor;
		cursor = cursor->next;
		call->buffer_head = cursor;
		rxrpc_check_buf(sent);
		sent->magic = RXGEN_BUF_DEAD;
		free(sent->buf);
		free(sent);
	} while (ret > 0);

	rxrpc_check_call(call);

	if (call->data_count > 0)
		goto more_to_send;
	if (call->data_count == 0 && !call->more_send)
		call->state++;

	if (!call->more_send) {
		if (call->state == rx_call_cl_encoding_params ||
		    call->state == rx_call_sv_encoding_response)
			call->state++;
	}

	if (call->state == rx_call_cl_waiting_for_response) {
		/* Prepare to decode the response */
		debug("Prep to decode\n");
		call->data_stop = call->data_cursor = call->data_start = call->buffer_head->buf;
		call->buffer_head->io_cursor = 0;
		call->data_count = 0;
	}

	return 0;
}

/*
 * Call back from the transport to indicate that data is ready.
 */
static void rxrpc_data_ready(struct kafs_call_handle *call_handle, void *caller_data)
{
	struct rx_call *call = caller_data;
	struct rx_buf *bufs[4] = { NULL }, *cursor;
	struct iovec iov[4];
	enum kafs_recv_condition cond;
	uint32_t supplementary_info;
	int i, ioc, ret;

	debug("Recv: buf[0]=%p data[0]=%p (io=%u)\n",
	      call->buffer_tail, call->buffer_tail->buf, call->buffer_tail->io_cursor);

	rxrpc_check_call(call);

	/* Set up some buffers */
	if (call->buffer_tail->io_cursor < RXGEN_BUFFER_SIZE) {
		bufs[0] = call->buffer_tail;
		iov[0].iov_base = bufs[0]->buf + bufs[0]->io_cursor;
		iov[0].iov_len = RXGEN_BUFFER_SIZE - bufs[0]->io_cursor;
		ioc = 1;
	} else {
		ioc = 0;
	}

	for (i = ioc; i < 4; i++) {
		bufs[i] = calloc(1, sizeof(struct rx_buf));
		bufs[i]->magic = RXGEN_BUF_MAGIC;
		if (!bufs[i]) {
			while (--i >= ioc) {
				bufs[i]->magic = RXGEN_BUF_DEAD;
				free(bufs[i]->buf);
				free(bufs[i]);
			}
			return;
		}
		bufs[i]->buf = malloc(RXGEN_BUFFER_SIZE);
		if (!bufs[i]->buf) {
			bufs[i]->magic = RXGEN_BUF_DEAD;
			free(bufs[i]);
			while (--i >= ioc) {
				bufs[i]->magic = RXGEN_BUF_DEAD;
				free(bufs[i]->buf);
				free(bufs[i]);
			}
			return;
		}
		iov[i].iov_base = bufs[i]->buf;
		iov[i].iov_len = RXGEN_BUFFER_SIZE;
	}
	ioc = 4;

	ret = transport->recv_data(call->handle, iov, ioc, &cond, &supplementary_info);
	debug("RECVDATA: %d\n", ret);
	if (ret == -1)
		return;

	call->bytes_received += ret;

	/* Attach any used buffers to the call and discard the rest */
	if (ret > 0) {
		for (i = 0; i < 4 && ret > 0; i++) {
			unsigned added = RXGEN_BUFFER_SIZE - bufs[i]->io_cursor;
			debug("xfer[%d] space=%u rem=%d\n", i, added, ret);
			if (added > (unsigned)ret)
				added = (unsigned)ret;
			bufs[i]->io_cursor += added;
			call->data_count += added;
			ret -= added;
			if (call->buffer_tail == bufs[i])
				continue;
			call->buffer_tail->next = bufs[i];
			call->buffer_tail = bufs[i];
			if (ret <= 0) {
				i++;
				break;
			}
		}

		for (; i < 4; i++) {
			bufs[i]->magic = RXGEN_BUF_DEAD;
			free(bufs[i]->buf);
			free(bufs[i]);
		}
	}

	rxrpc_check_call(call);

	for (cursor = call->buffer_head; cursor; cursor = cursor->next)
		debug("RecvQ buf=%p data=%p iocur=%u\n",
		      cursor, cursor->buf, cursor->io_cursor);

	/* Process the call state given to us by the receiver. */
	switch (cond) {
	case kafs_recv_more:
		break;
	case kafs_recv_final_ack:
		if (call->state == rx_call_sv_waiting_for_final_ack) {
			call->state = rx_call_sv_complete;
		} else {
			fprintf(stderr, "RxRPC: Recv-Ack in bad call state (%d)\n",
				call->state);
			abort();
		}
	case kafs_recv_complete:
		call->known_to_kernel = 0;
	case kafs_recv_complete_now_send:
		call->more_recv = false;
		break;

	case kafs_recv_abort:
		call->conn->last_abort_code = supplementary_info;
		call->abort_code = supplementary_info;
		call->error_code = ECONNABORTED;
		call->state = rx_call_remotely_aborted;
		call->known_to_kernel = 0;
		call->more_recv = false;
		break;
	case kafs_recv_net_error:
		call->error_code = supplementary_info;
		call->state = rx_call_net_error;
		call->known_to_kernel = 0;
		call->more_recv = false;
		break;
	case kafs_recv_local_error:
		call->error_code = supplementary_info;
		call->state = rx_call_local_error;
		call->known_to_kernel = 0;
		call->more_recv = false;
		break;
	case kafs_recv_busy:
		call->error_code = ECONNREFUSED;
		call->state = rx_call_rejected_busy;
		call->known_to_kernel = 0;
		call->more_recv = false;
		break;
	default:
		abort();
	}

	/* Switch into appropriate decode state */
loop:
	switch (call->state) {
	case rx_call_cl_waiting_for_response:
	case rx_call_sv_waiting_for_opcode:
		call->state++;
	case rx_call_cl_decoding_response:
	case rx_call_sv_decoding_opcode:
	case rx_call_sv_decoding_params:
		if (call->need_size == UINT_MAX) {
			/* A split mode call wants everything as it comes in,
			 * and is willing to settle for nothing also.
			 */
			if (call->data_count <= 0 && call->more_recv)
				return;
		} else if (call->data_count < call->need_size) {
			if (!call->more_recv) {
				/* Short data */
				debug("SHORT %u < %x\n", call->data_count, call->need_size);
				call->error_code = ENODATA;
				rxrpc_abort_call(call, 1);
			}
			return;
		}

		/* Handle reception of the opcode in a server-side call.  Note
		 * that a call might have no parameters, in which case we
		 * should've got !MSG_MORE and incremented the state again.
		 */
		if (call->state == rx_call_sv_decoding_opcode) {
			fprintf(stderr, "Need to determine service based on opcode\n");
			abort();
			call->state++;
			goto loop;
		}

		debug("DEC %u > %u\n", call->data_count, call->need_size);
		ret = call->decoder(call);
		if (ret < 0) {
			rxrpc_abort_call(call, 1);
			break;
		}

		if (ret == 1)
			goto loop; /* The decoder wants more data */

		call->state++;
	case rx_call_cl_wait_for_no_MSG_MORE:
	case rx_call_sv_wait_for_no_MSG_MORE:
		rxgen_dec_discard_excess(call);
		if (!call->more_recv) {
			call->state++;
			if (call->state == rx_call_sv_processing) {
				/* Prepare to encode the response */
				call->data_cursor = call->data_start = call->buffer_head->buf;
				call->data_stop = call->buffer_head->buf + RXGEN_BUFFER_SIZE;
				call->buffer_head->io_cursor = 0;
				call->data_count = 0;
				call->buffer_space = 0;

				/* Invoke the processor */
				call->processor(call);
				return;
			}
		}
		break;

	case rx_call_cl_complete:
	case rx_call_sv_complete ... rx_call_rejected_busy:
		break;

	case rx_call_sv_processing:
		call->failed(call);
		return;

	default:
		fprintf(stderr, "RxRPC: Recv in bad call state (%d)\n", call->state);
		abort();
	}
}

/*
 * Abort a call.
 */
void rxrpc_abort_call(struct rx_call *call, uint32_t abort_code)
{

	if (call->known_to_kernel) {
		transport->abort_call(call->handle, abort_code);
		call->known_to_kernel = 0;
	}
	call->state = rx_call_locally_aborted;
}

/*
 * Terminate a call, aborting it if necessary.
 */
void rxrpc_terminate_call(struct rx_call *call, uint32_t abort_code)
{
	struct rx_buf *cursor, *next;

	rxrpc_check_call(call);
	if (call->known_to_kernel)
		rxrpc_abort_call(call, abort_code);
	transport->terminate_call(call->handle);
	call->magic = 0x7a7b7c7d;
	for (cursor = call->buffer_head; cursor; cursor = next) {
		rxrpc_check_buf(cursor);
		next = cursor->next;
		cursor->magic = RXGEN_BUF_DEAD;
		free(cursor->buf);
		free(cursor);
	}
	if (call->decoder_cleanup)
		call->decoder_cleanup(call);
	free(call);
}

/*
 * Poll the transport and receive data if there's anything available.
 */
int rxrpc_maybe_recv_data(struct rx_connection *conn)
{
	return transport->poll_connection(conn->handle, true);
}

/*
 * Run a single call synchronously.
 */
int rxrpc_run_sync_call(struct rx_call *call)
{
	int ret;

	while (call->known_to_kernel) {
		debug("Run call %d\n", call->known_to_kernel);
		ret = transport->poll_connection(call->conn->handle, false);
		if (ret < 0)
			return ret;
	}

	switch (call->state) {
	case rx_call_cl_complete:
		debug("Call complete\n");
		return 0;
	case rx_call_remotely_aborted ... rx_call_rejected_busy:
		debug("Call failed\n");
		errno = call->error_code;
		return -1;
	default:
		fprintf(stderr, "RxRPC: Ended in bad call state (%d)\n", call->state);
		abort();
	}
}

/*
 * Discard excess received data
 */
int rxgen_dec_discard_excess(struct rx_call *call)
{
	struct rx_buf *spent, *cursor;
	unsigned io_cursor;

	rxrpc_check_call(call);

	while (cursor = call->buffer_head,
	       cursor != call->buffer_tail
	       ) {
		spent = cursor;
		call->buffer_head = cursor->next;
		spent->magic = RXGEN_BUF_DEAD;
		free(spent->buf);
		free(spent);
	}

	rxrpc_check_call(call);

	io_cursor = cursor->io_cursor;
	call->data_start = call->data_cursor = call->data_stop = cursor->buf + io_cursor;
	call->data_count = 0;

	if (__builtin_expect(call->error_code, 0))
		return -1;
	return 0;
}

/*
 * Encode a string of bytes
 */
int rxrpc_enc_blob(struct rx_call *call, const void *data, size_t size)
{
	struct rx_buf *cursor, *new;
	uint8_t *buf;
	size_t seg;

	rxrpc_check_call(call);

	if (size == 0)
		return 0;
	if (call->error_code)
		return -1;

	for (;;) {
		if (call->data_cursor < call->data_stop) {
			seg = call->data_stop - call->data_cursor;
			if (seg > size)
				seg = size;
			memcpy(call->data_cursor, data, seg);
			data += seg;
			size -= seg;
			call->data_cursor += seg;
			if (size <= 0)
				return 0;
		}

		rxrpc_check_call(call);

		if (rxrpc_post_enc(call) < 0)
			return -1;

		new = calloc(1, sizeof(struct rx_buf));
		if (!new)
			goto handle_oom;
		new->buf = buf = malloc(RXGEN_BUFFER_SIZE);
		if (!buf)
			goto handle_oom_buf;

		new->magic = RXGEN_BUF_MAGIC;
		cursor = call->buffer_tail;
		cursor->next = new;
		call->data_cursor = call->data_start = buf;
		call->data_stop = buf + RXGEN_BUFFER_SIZE;
		call->buffer_space += RXGEN_BUFFER_SIZE;
		call->buffer_tail = new;
	}

handle_oom_buf:
	free(new);
handle_oom:
	call->error_code = ENOMEM;
	return -1;
}

/*
 * Slow path for decoding from data read from the buffer list.
 */
void rxrpc_enc_slow(struct rx_call *call, net_xdr_t data)
{
	rxrpc_enc_blob(call, &data, sizeof(data));
}

/*
 * Decode a blob.
 *
 * This works progressively and so may need to be called multiple times for any
 * particular set.
 *
 * call->blob_size indicates how many bytes we need to extract in total and
 * call->blob_offset indicates how many we have done so far.  blob_offset is
 * updated before returning.
 */
void rxrpc_dec_blob(struct rx_call *call)
{
	unsigned needed, segment;

	rxrpc_check_call(call);
	rxrpc_post_dec(call);

	needed = call->blob_size - call->blob_offset;
	segment = call->data_stop - call->data_cursor;
	debug("DEC BLOB dc=%u bsize=%u seg=%u\n", call->data_count, needed, segment);

	if (segment > 0) {
		if (segment > needed)
			segment = needed;

		debug("DEC BLOB copy %u\n", segment);
		memcpy(call->blob + call->blob_offset, call->data_cursor, segment);
		call->blob_decoded += segment;
		call->blob_offset += segment;
		call->data_cursor += segment;
		if (call->blob_offset >= call->blob_size)
			return;
	}

	rxrpc_dec_advance_buffer(call);
}

/*
 * Progress the buffer chain during decoding.
 */
void rxrpc_dec_advance_buffer(struct rx_call *call)
{
	struct rx_buf *cursor, *spent;
	unsigned segment;
	uint8_t *new_stop;

	/* More data may have come into a partially filled buffer from which we
	 * previously read.
	 */
	cursor = call->buffer_head;
	segment = cursor->io_cursor;
	debug("DEC Advance %u\n", segment);
	new_stop = cursor->buf + segment;
	if (call->data_stop >= new_stop) {
		/* This buffer must then be completely used as we're required to check
		 * amount received before reading it.
		 */
		if (call->data_stop > new_stop) {
			debug("buffer overrun (%p > %p)\n", call->data_stop, new_stop);
			abort();
		}
		if (new_stop != cursor->buf + cursor->io_cursor) {
			debug("buffer not consumed (%p != %p)\n",
			      new_stop, cursor->buf + cursor->io_cursor);
			abort();
		}
		if (call->buffer_tail == cursor) {
			debug("unexpectedly out of data\n");
			abort();
		}

		/* Move to the next buffer */
		rxrpc_post_dec(call);
		spent = cursor;
		cursor = cursor->next;
		call->buffer_head = cursor;

		call->data_cursor = call->data_start = cursor->buf;
		segment = cursor->io_cursor;
		if (segment == 0)
			abort();
		new_stop = cursor->buf + segment;

		spent->magic = RXGEN_BUF_DEAD;
		free(spent->buf);
		free(spent);
	}

	call->data_stop = new_stop;
	rxrpc_check_call(call);
}

/*
 * Slow path for decoding a 4-byte integer read from the buffer list.
 */
uint32_t rxrpc_dec_slow(struct rx_call *call)
{
	net_xdr_t x;
	uint32_t ret;

	debug("DEC SLOW %lu %ld\n",
	      (unsigned long)call->data_cursor & 3,
	      call->data_cursor - call->data_stop);

	call->blob_size = 4;
	call->blob_offset = 0;
	call->blob = &x;

	/* We should only come here if there is sufficient data in the
	 * buffers for us to complete all the copy operation(s).
	 */
	if (call->data_count < call->blob_size)
		abort();

	while (call->blob_offset < call->blob_size)
		rxrpc_dec_blob(call);

	ret = ntohl(x);
	debug("DEC SLOW = %x\n", ret);
	return ret;
}
